/*----------------------------------------------------------------------------
miniAudicle
Cocoa GUI to chuck audio programming environment

Copyright (c) 2005 Spencer Salazar.  All rights reserved.
http://chuck.cs.princeton.edu/
http://soundlab.cs.princeton.edu/

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
U.S.A.
-----------------------------------------------------------------------------*/

//-----------------------------------------------------------------------------
// file: NumberedTextView.mm
// desc: view class for implementing a text view with scrollers and line numbers
//       Based on Koen van der Drift's MyTextView class.  
//
// author: Spencer Salazar (ssalazar@cs.princeton.edu)
// date: Spring 2006
//-----------------------------------------------------------------------------

#import "NumberedTextView.h"
#import "miniAudiclePreferencesController.h"

#define MARGIN_WIDTH 30

#define USE_NEW_SYNTAX_COLORING 0

@implementation NumberedTextView

-(id)initWithFrame:(NSRect)frame
{
    if(self = [super initWithFrame:frame])
    {
		[self initLineMargin:frame];
        [self initSyntaxHighlighting];
        open_bracket = NO;
        close_bracket = -1;
        use_tabs = TRUE;
        tab_width = 4;
    }
	
    return self;
}

- (void)initSyntaxHighlighting
{    
    syntax_highlighter = nil;
    colorer = nil;
    //enable_syntax_highlighting = [[NSUserDefaults standardUserDefaults] boolForKey:
#if !defined( USE_NEW_SYNTAX_COLORING ) || !USE_NEW_SYNTAX_COLORING
    [[NSNotificationCenter defaultCenter] addObserver:self
                                             selector:@selector(textStorageDidProcessEditing:)
                                                 name:NSTextStorageDidProcessEditingNotification
                                               object:[text_view textStorage]];
    
    [[NSNotificationCenter defaultCenter] addObserver:self
                                             selector:@selector(syntaxColoringChanged:)
                                                 name:mASyntaxColoringChangedNotification
                                               object:nil];
#endif
    //[text_view setDelegate:self];
    
    NSString * bg_color_hex = [[[NSUserDefaults standardUserDefaults] dictionaryForKey:IDEKit_TextColorsPrefKey] objectForKey:IDEKit_NameForColor( IDEKit_kLangColor_Background )];
    if( bg_color_hex )
    {
        NSColor * bg_color = [NSColor colorWithHTML:bg_color_hex];
        [text_view setBackgroundColor:bg_color];
        if( [bg_color brightnessComponent] < .5 )
            [text_view setInsertionPointColor:[NSColor whiteColor]];
        if( [bg_color brightnessComponent] >= .5 )
            [text_view setInsertionPointColor:[NSColor blackColor]];
    }
    
#if defined( USE_NEW_SYNTAX_COLORING ) && USE_NEW_SYNTAX_COLORING
    mASyntaxHighlighter * mash = [[mASyntaxHighlighter alloc] initWithTextStorage:[text_view textStorage]];
#endif
}

- (void)setSyntaxHighlighter:(IDEKit_LexParser *)sh
                     colorer:(id)c
{
    syntax_highlighter = [sh retain];
    colorer = nil;//c;
}

- (void)textStorageDidProcessEditing:(NSNotification *)notification
{
    if( !syntax_highlighter )
        return;
    
    NSTextStorage * ts = [text_view textStorage];
    if ( [ts editedMask] == NSTextStorageEditedAttributes )
        return; 
    NSRange editedRange = [ts editedRange];
    NSString * text = [ts string];
    unsigned int startIndex, lineEndIndex, contentsEndIndex;
    [text getLineStart:&startIndex end:&lineEndIndex 
           contentsEnd:&contentsEndIndex forRange:editedRange];
    [syntax_highlighter colorString:ts 
                              range:NSMakeRange( startIndex, contentsEndIndex - startIndex ) 
                            colorer:colorer];
}

- (void)syntaxColoringChanged:(NSNotification *)n
{
    if( !syntax_highlighter )
        return;
    
    NSTextStorage * ts = [text_view textStorage];
    if( [ts length] > 0 )
        [syntax_highlighter colorString:ts 
                                  range:NSMakeRange( 0, [ts length] ) 
                                colorer:colorer];
    
    NSString * bg_color_hex = [[[NSUserDefaults standardUserDefaults] dictionaryForKey:IDEKit_TextColorsPrefKey] objectForKey:IDEKit_NameForColor( IDEKit_kLangColor_Background )];
    if( bg_color_hex )
    {
        NSColor * bg_color = [NSColor colorWithHTML:bg_color_hex];
        [text_view setBackgroundColor:bg_color];
        if( [bg_color brightnessComponent] < .5 )
            [text_view setInsertionPointColor:[NSColor whiteColor]];
        if( [bg_color brightnessComponent] >= .5 )
            [text_view setInsertionPointColor:[NSColor blackColor]];
    }
}

- (BOOL)textView:(NSTextView *)aTextView doCommandBySelector:(SEL)aSelector
{
    if( aSelector == @selector(insertNewline:) )
    {
        [text_view insertNewline:self];
        
        // handle tab level auto indentation
        NSTextStorage * ts = [text_view textStorage];
        NSString * text = [ts string];
        NSRange range = [text_view selectedRange];
        unsigned line_start, tab_level_end, line_end = range.location - 1;

        // first find the beginning of the preceding line
        range = [text rangeOfCharacterFromSet:[NSCharacterSet characterSetWithCharactersInString:@"\n\r"]
                                      options:NSBackwardsSearch
                                        range:NSMakeRange( 0, line_end )];
        line_start = ( range.location == NSNotFound ? 0 : range.location + 1 );
        
        // now find the first non-whitespace character on the line
        range = [text rangeOfCharacterFromSet:[[NSCharacterSet whitespaceCharacterSet] invertedSet]
                                      options:0
                                        range:NSMakeRange( line_start, line_end - line_start + 1)];
        // since there is a \n at line_end, we have to find something...
        // assert( range.location != NSNotFound );
        tab_level_end = range.location;
        
        // now that we have the start of the line and the end of the opening
        // whitespace, we can just insert the opening whitespace at the
        // beginning of the next line
        [ts insertAttributedString:[ts attributedSubstringFromRange:NSMakeRange( line_start, tab_level_end - line_start )]
                           atIndex:line_end + 1];
        
        if( open_bracket )
        {
            [text_view doCommandBySelector:@selector(insertTab:)];
        }
        
        return YES;
    }
    
    if( aSelector == @selector(insertTab:) )
    {
        [text_view insertTab:self];
        
        return YES;
    }
    
    return NO;
}

- (BOOL)textView:(NSTextView *)aTextView shouldChangeTextInRange:(NSRange)affectedCharRange replacementString:(NSString *)replacementString
{
    changed_range = NSMakeRange( affectedCharRange.location, [replacementString length] );
    
    return YES;
}

- (void)textDidEndEditing:(NSNotification *)n
{
    /*
    NSString * text = [text_view string];
    NSRange range = [text rangeOfCharacterSet:[NSCharacterSet characterSetFromCharactersInString:@"{\n}"]
                                      options:0
                                        range:changed_range];
    
    */
}

- (void)initLineMargin:(NSRect) frame
{
    NSRect r = NSMakeRect( MARGIN_WIDTH, 0, frame.size.width - MARGIN_WIDTH, frame.size.height );

    // initialize a scrolling view
	NSScrollView * scroll_view = [[[NSScrollView alloc] initWithFrame:r] autorelease];
	
	// set scroll view properties
	[scroll_view setHasVerticalScroller:YES];
	[scroll_view setHasHorizontalScroller:YES];
	[scroll_view setAutoresizingMask:NSViewWidthSizable | NSViewHeightSizable];
	
	NSRect r2 = { { 0, 0 }, [scroll_view contentSize] };
		
	// create a text view, adjust it for display
	text_view = [[NSTextView alloc] initWithFrame:r2];
	
	// set various text view properties
	[text_view setEditable:YES];
	[text_view setAllowsUndo:YES];
	[text_view setUsesFindPanel:YES];
	[text_view setHorizontallyResizable:YES];
	[text_view setMaxSize:NSMakeSize(FLT_MAX, FLT_MAX)];
	[text_view setAutoresizingMask:NSViewWidthSizable | NSViewHeightSizable];
    [text_view setSmartInsertDeleteEnabled:NO];
	[[text_view textContainer] setWidthTracksTextView:NO];
	[[text_view textContainer] setContainerSize:NSMakeSize(FLT_MAX, FLT_MAX)];
	
    NSUnarchiver * uar = [[[NSUnarchiver alloc] initForReadingWithData:[[NSUserDefaults standardUserDefaults] objectForKey:mAPreferencesDefaultFont]] autorelease];
	[text_view setFont:[[[NSFont alloc] initWithCoder:uar] autorelease]];

    [scroll_view setDocumentView:text_view];
    
    [self addSubview:scroll_view];
    
    // listen to updates from the window to force a redraw - eg when the window resizes.

	[[NSNotificationCenter defaultCenter] addObserver:self selector:@selector(windowDidUpdate:)
						name:NSWindowDidUpdateNotification object:[self window]];
	
	marginAttributes = [[NSMutableDictionary alloc] init];
    
	[marginAttributes setObject:[NSFont boldSystemFontOfSize:8] forKey: NSFontAttributeName];
	[marginAttributes setObject:[NSColor darkGrayColor] forKey: NSForegroundColorAttributeName];

	drawNumbersInMargin = YES;
	drawLineNumbers = YES;
}

- (void)dealloc
{
    [[NSNotificationCenter defaultCenter] removeObserver:self];
    [marginAttributes release];
    if( syntax_highlighter )
        [syntax_highlighter release];

    [super dealloc];
}

- (void)drawRect:(NSRect)aRect
{
    [super drawRect:aRect];
    
    [self drawEmptyMargin: [self marginRect]];
    
    if ( drawNumbersInMargin )
    {
        [self drawNumbersInMargin: [self marginRect]];
    }
}


- (void)windowDidUpdate:(NSNotification *)notification
{
    [self updateMargin];
}

- (void)updateLayout
{
    [self updateMargin];
}


-(void)updateMargin
{
    [self setNeedsDisplayInRect:[self marginRect]];
}

-(NSRect)marginRect
{
    NSRect  r;
    
    r = [self bounds];
    r.size.width = MARGIN_WIDTH;

    return r;
}

-(void)drawEmptyMargin:(NSRect)aRect
{
    /*
     These values control the color of our margin. Giving the rect the 'clear' 
     background color is accomplished using the windowBackgroundColor.  Change 
     the color here to anything you like to alter margin contents.
     */
    [[NSColor controlColor] set];
    [NSBezierPath fillRect: aRect]; 
    
    // These points should be set to the left margin width.
    NSPoint top = NSMakePoint(aRect.size.width, [self bounds].size.height);
    NSPoint bottom = NSMakePoint(aRect.size.width, 0);
    
    // This draws the dark line separating the margin from the text area.
    [[NSColor grayColor] set];
    [NSBezierPath setDefaultLineWidth:0.75];
    [NSBezierPath strokeLineFromPoint:top toPoint:bottom];
}

-(void) drawNumbersInMargin:(NSRect)aRect;
{
	UInt32 index, lineNumber;
	NSRange	lineRange;
	NSRect lineRect;
    
	NSLayoutManager * layoutManager = [text_view layoutManager];
	NSTextContainer * textContainer = [text_view textContainer];

    if( !layoutManager || !textContainer )
        exit( 0 );
    
	// Only get the visible part of the scroller view
	NSRect documentVisibleRect = [[text_view enclosingScrollView] documentVisibleRect];

	// Find the glyph range for the visible glyphs
	NSRange glyphRange = [layoutManager glyphRangeForBoundingRect: documentVisibleRect inTextContainer: textContainer];

	// Calculate the start and end indexes for the glyphs	
	unsigned start_index = glyphRange.location;
	unsigned end_index = glyphRange.location + glyphRange.length;
 
	index = 0;
	lineNumber = 1;

	// Skip all lines that are visible at the top of the text view (if any)
	while (index < start_index)
	{
		lineRect = [layoutManager lineFragmentRectForGlyphAtIndex:index effectiveRange:&lineRange];
		index = NSMaxRange( lineRange );
		++lineNumber;
	}

	for ( index = start_index; index < end_index; lineNumber++ )
	{
		lineRect = [layoutManager lineFragmentRectForGlyphAtIndex:index effectiveRange:&lineRange];
		index = NSMaxRange( lineRange );
        [self drawOneNumberInMargin:lineNumber inRect:lineRect];
    }
    
    lineRect = [layoutManager extraLineFragmentRect];
    if( lineRect.size.width != 0 && lineRect.size.height != 0 )
        [self drawOneNumberInMargin:lineNumber inRect:lineRect];
}


-(void)drawOneNumberInMargin:(unsigned) aNumber inRect:(NSRect)r
{
    NSString * s;
    NSSize stringSize;
    NSRect visible_rect = [[text_view enclosingScrollView] documentVisibleRect];
    
    s = [NSString stringWithFormat:@"%d", aNumber, nil];
    stringSize = [s sizeWithAttributes:marginAttributes];
    
    NSPoint p = NSMakePoint( MARGIN_WIDTH - stringSize.width - 1.0, 
                             [self frame].size.height - r.origin.y - 
                             r.size.height / 2 - stringSize.height / 3 + 
                             visible_rect.origin.y );
    
    [s drawAtPoint:p withAttributes:marginAttributes];
}


- ( NSTextView * )textView
{
    return text_view;
}

@end
