/*----------------------------------------------------------------------------
miniAudicle
Cocoa GUI to chuck audio programming environment

Copyright (c) 2005 Spencer Salazar.  All rights reserved.
http://chuck.cs.princeton.edu/
http://soundlab.cs.princeton.edu/

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
U.S.A.
-----------------------------------------------------------------------------*/

//-----------------------------------------------------------------------------
// file: miniAudicleVMMonitor.mm
// desc: controller class for floating Virtual Machine monitor panel
//
// author: Spencer Salazar (ssalazar@princeton.edu)
// date: Autumn 2005
//-----------------------------------------------------------------------------

#import "miniAudicleVMMonitor.h"
#import "miniAudicleController.h"

const static NSString * shred_column_id = @"1";
const static NSString * name_column_id = @"2";
const static NSString * time_column_id = @"3";
const static NSString * remove_column_id = @"4";

int compare_shred_vectors( const vector< Chuck_VM_Shred_Status * > & a,
                           const vector< Chuck_VM_Shred_Status * > & b )
/* quickly determine if the two vectors are equal */
{
    vector< Chuck_VM_Shred_Status * >::size_type i, 
        lenA = a.size(), lenB = b.size();

    if( lenA != lenB )
        return 1;
        
    if( lenA == 0 )
        return 0;
    
    Chuck_VM_Shred_Status * cvmssA, * cvmssB;
    
    for( i = 0; i < lenA; i++ )
    {
        cvmssA = a[i];
        cvmssB = b[i];
        
        if( cvmssA->xid != cvmssB->xid ||
            cvmssA->start != cvmssB->start )
            /* a shred is uniquely defined by ( shred id, start time ) */
        {
            return 1;
        }
    }
    
    return 0;
}

@implementation miniAudicleVMMonitor

- (id)init
{
	if( self = [super init] )
	{
        status_buffers = new Chuck_VM_Status[2];
        omd = new vector< mAShredOptimizationMetadata >( 30 );
        most_recent_status = status_buffers;
        which_status_buffer = 0;
        [running_time_text setStringValue:@""];
        [shreds_text setStringValue:@""];
	}
	
	return self;
}

- (void)awakeFromNib
{
    [[shred_table headerView] setMenu:header_menu];
    shred_column = [[shred_table tableColumnWithIdentifier:@"shred"] retain];
    name_column = [[shred_table tableColumnWithIdentifier:@"name"] retain];
    time_column = [[shred_table tableColumnWithIdentifier:@"time"] retain];
    remove_column = [[shred_table tableColumnWithIdentifier:@"remove"] retain];
    
    [shred_column setIdentifier:shred_column_id];
    [name_column setIdentifier:name_column_id];
    [time_column setIdentifier:time_column_id];
    [remove_column setIdentifier:remove_column_id];
}

- (void)activateMonitor:(id)sender
{
	if( [panel isKeyWindow] )
		[panel close];
	
	else
		[panel makeKeyAndOrderFront:sender];
}

- (void)update:(id)data
{
    most_recent_status = status_buffers + which_status_buffer;
    which_status_buffer = ( which_status_buffer + 1 ) % 2;
    [controller miniAudicle]->status( most_recent_status );
    if( start_time == 0 )
        start_time = time( NULL ) - (time_t)(most_recent_status->now_system / most_recent_status->srate);
    
    if( compare_shred_vectors( status_buffers[0].list, status_buffers[1].list ) )
    {
        [shred_table reloadData];
        [shreds_text setStringValue:[NSString stringWithFormat:@"%u", 
            most_recent_status->list.size()]];
    }
    
    else
    {
        [shred_table setNeedsDisplayInRect:[shred_table rectOfColumn:[shred_table columnWithIdentifier:time_column_id]]];
    }
	
    
	time_t current_time = time( NULL ) - start_time;
    if( current_time != last_time )
    {
        [running_time_text setStringValue:[NSString stringWithFormat:@"%u:%.2u", 
            current_time / 60, current_time % 60 ]];
    }
    
    last_time = current_time;
}

- (void)toggleVM:(id)sender
{
	[controller toggleVM:sender];
}

- (void)vm_on
// TODO: account for vm_on failure
{
	[vm_toggle_button setTitle:@"Stop Virtual Machine"];
	vm_on = YES;

	timer = [NSTimer timerWithTimeInterval:0.1
									target:self 
								  selector:@selector(update:)
								  userInfo:nil
								   repeats:YES];
	[timer retain];
	NSRunLoop * run_loop = [NSRunLoop currentRunLoop];
	[run_loop addTimer:timer forMode:NSDefaultRunLoopMode];	
	[controller miniAudicle]->status( most_recent_status );
	start_time = 0;
    last_time = -1;
}

- (void)vm_off
{
	[timer invalidate];
	[timer release];
	[vm_toggle_button setTitle:@"Start Virtual Machine"];
	vm_on = NO;
	most_recent_status->clear();
	[shred_table reloadData];
	[running_time_text setStringValue:@""];
	[shreds_text setStringValue:@""];
}

// table data source functions
- (int)numberOfRowsInTableView:(NSTableView *)table_view
{
	if( vm_on )
    {
        int s = most_recent_status->list.size();
        if( s >= omd->size() )
            omd->resize( s * 2 );
        
		return s;
    }
	
	else
		return 0;
}

- (id)tableView:(NSTableView *)table_view 
objectValueForTableColumn:(NSTableColumn *)table_column
			row:(int)rowIndex
{
	if( [table_column identifier] == shred_column_id )
		return [NSNumber numberWithInt:most_recent_status->list[rowIndex]->xid];
	
	else if( [table_column identifier] == name_column_id )
		return [NSString stringWithCString:most_recent_status->list[rowIndex]->name.c_str()];
	
	else if( [table_column identifier] == time_column_id )
	{
        //double shred_running_time = ((double)time( NULL )) - ((double)start_time) - ( most_recent_status->list[rowIndex]->start / most_recent_status->srate );
        time_t shred_running_time = rint(((double)time( NULL )) - ((double)start_time) - ( most_recent_status->list[rowIndex]->start / most_recent_status->srate ));
        if( shred_running_time < 0 )
            shred_running_time = 0;
		
        if( rowIndex < omd->size() && shred_running_time != (*omd)[rowIndex].last_time )
        {
            [(*omd)[rowIndex].last_string autorelease];
            //(*omd)[rowIndex].last_string = [[NSString stringWithFormat:@"%.0f:%02.0f", floor( shred_running_time / 60.0 ), fmod( shred_running_time, 60.0 )] retain];
            (*omd)[rowIndex].last_string = [[NSString stringWithFormat:@"%u:%02u", shred_running_time / 60, shred_running_time % 60] retain];
            (*omd)[rowIndex].last_time = shred_running_time;
        }
        
        return (*omd)[rowIndex].last_string;
	}
	
	else if( [table_column identifier] == remove_column_id )//*/
		return @"-";
}

- (void)removeShred:(id)sender
{
	string result;
	[controller miniAudicle]->remove_shred( most_recent_status->list[[sender clickedRow]]->xid, result );
}

- (void)removeShredTableColumn:(id)sender
{
    if( [sender tag] == 0 )
    {
        if( [sender state] == NSOnState )
        {
            shred_column_number = [shred_table columnWithIdentifier:@"shred"];
            [shred_table removeTableColumn:[shred_table tableColumnWithIdentifier:@"shred"]];
            [sender setState:NSOffState];
        }
        
        else
        {
            [shred_table addTableColumn:shred_column];
            [shred_table moveColumn:[shred_table columnWithIdentifier:@"shred"] toColumn:shred_column_number];
            [sender setState:NSOnState];
        }
    }
    
    else if( [sender tag] == 1 )
    {
        if( [sender state] == NSOnState )
        {
            name_column_number = [shred_table columnWithIdentifier:@"name"];
            [shred_table removeTableColumn:[shred_table tableColumnWithIdentifier:@"name"]];
            [sender setState:NSOffState];
        }
        
        else
        {
            [shred_table addTableColumn:name_column];
            [shred_table moveColumn:[shred_table columnWithIdentifier:@"name"] toColumn:name_column_number];
            [sender setState:NSOnState];
        }
    }

    else if( [sender tag] == 2 )
    {
        if( [sender state] == NSOnState )
        {
            time_column_number = [shred_table columnWithIdentifier:@"time"];
            [shred_table removeTableColumn:[shred_table tableColumnWithIdentifier:@"time"]];
            [sender setState:NSOffState];
        }
        
        else
        {
            [shred_table addTableColumn:time_column];
            [shred_table moveColumn:[shred_table columnWithIdentifier:@"time"] toColumn:time_column_number];
            [sender setState:NSOnState];
        }
    }

    else if( [sender tag] == 3 )
    {
        if( [sender state] == NSOnState )
        {
            remove_column_number = [shred_table columnWithIdentifier:@"remove"];
            [shred_table removeTableColumn:[shred_table tableColumnWithIdentifier:@"remove"]];
            [sender setState:NSOffState];
        }
        
        else
        {
            [shred_table addTableColumn:remove_column];
            [shred_table moveColumn:[shred_table columnWithIdentifier:@"remove"] toColumn:remove_column_number];
            [sender setState:NSOnState];
        }
    }
}

@end
