/*----------------------------------------------------------------------------
miniAudicle
Cocoa GUI to chuck audio programming environment

Copyright (c) 2005 Spencer Salazar.  All rights reserved.
http://chuck.cs.princeton.edu/
http://soundlab.cs.princeton.edu/

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
U.S.A.
-----------------------------------------------------------------------------*/

//-----------------------------------------------------------------------------
// file: miniAudicle_ui_elements.h
// desc: for supporting UI controls created and used in ChucK code
//
// author: Spencer Salazar (ssalazar@cs.princeton.edu)
// date: Autumn 2005
//-----------------------------------------------------------------------------
#import <Cocoa/Cocoa.h>

#include "miniAudicle_ui_elements.h"
using namespace std;
//using namespace miniAudicle::UI;
using namespace UI;

//------------------------------------------------------------------------------
// name: mAFlippedView
// desc: NSView with flipped (traditional) Y axis
//------------------------------------------------------------------------------
@interface mAFlippedView : NSView
- (BOOL)isFlipped;
@end

@implementation mAFlippedView

- (BOOL)isFlipped
{
    return YES;
}

@end

//------------------------------------------------------------------------------
// name: mAUIElement
// desc: ...
//------------------------------------------------------------------------------
@interface mAUIElement : NSObject
{
    Element * owner;
    NSView * master_view;
}

- (void)link:( Element * )e;
- (id)init;
- (void)display;
- (void)hide;
- (void)destroy;
- (void)setSize;
- (void)setPosition;

- (void)removeParent;

- (NSView * )masterView;
- (void)setMasterView:(NSView *)v;

@end

@implementation mAUIElement

- (void)link:( Element * )e
{
    owner = e;    
}

- (id)init
{
    return [super init];
}

- (void)display
{
    if( master_view )
        [master_view setHidden:NO];
}

- (void)hide
{
    if( master_view )
        [master_view setHidden:YES];
}

- (void)destroy
{
    if( master_view )
    {
        [master_view release];
        master_view = NULL;
    }
    
    [self autorelease];
}

- (void)setSize
{
    if( master_view )
        [master_view setFrameSize:NSMakeSize( owner->get_width(), owner->get_height() )];
}

- (void)setPosition
{
    if( master_view )
        [master_view setFrameOrigin:NSMakePoint( owner->get_x(), owner->get_y() )];
}

- (void)removeParent
{
    [master_view removeFromSuperview];
}

- (NSView * )masterView
{
    return master_view;
}

- (void)setMasterView:(NSView *)v
{
    master_view = v;
}

@end

//------------------------------------------------------------------------------
// name: mAUIView
// desc: ...
//------------------------------------------------------------------------------
@interface mAUIView : mAUIElement
{
    NSWindow * panel;
}

- (id)init;
- (void)display;
- (void)hide;
- (void)destroy;
- (void)setName;
- (void)setSize;
- (void)setPosition;
- (void)addSubview:(NSView *)v;
- (void)addElement:(mAUIElement *)e;

@end

@implementation mAUIView

- (id)init
{
    if( self = [super init] )
    {
        panel = [[NSWindow alloc] initWithContentRect:NSMakeRect( 0, 0, 500, 500 )
                                            styleMask:NSTitledWindowMask | NSMiniaturizableWindowMask | NSClosableWindowMask
                                              backing:NSBackingStoreBuffered
                                                defer:YES];
        [panel setReleasedWhenClosed:NO];
        master_view = [[[mAFlippedView alloc] init] autorelease];
        [panel setContentView:master_view];
        [panel center];
    }
    
    return self;
}

- (void)display
{
    [panel makeKeyAndOrderFront:nil];
}

- (void)hide
{
    [panel close];
}

- (void)destroy
{
    [panel close];
    [panel release];
    [self autorelease];
}

- (void)setName
{
    [panel setTitle:[NSString stringWithCString:(owner->get_name()).c_str()]];
}

- (void)setSize
{
    [panel setContentSize:NSMakeSize( owner->get_width(), owner->get_height() )];
}

- (void)setPosition
{
    [panel setFrameOrigin:NSMakePoint( owner->get_x(), owner->get_y() )];
}

- (void)addSubview:(NSView *)v
{
    [[panel contentView] addSubview:v];
}

- (void)addElement:(mAUIElement *)e
{
    [[panel contentView] addSubview:[e masterView]];
}

@end

//------------------------------------------------------------------------------
// name: mAUISlider
// desc: ...
//------------------------------------------------------------------------------
@interface mAUISlider : mAUIElement
{
    Slider * s_owner;
    NSSlider * slider;
    NSTextField * title;
	NSTextField * value;
    NSString * display_format;
}

- (void)link:( Slider * )s;
- (id)init;
- (void)destroy;
- (void)setName;
- (void)setValue;
- (void)setRange;
- (void)setDisplayFormat;
- (void)updateDisplay;
- (void)sliderDidChange;

@end

@implementation mAUISlider

- (void)link:( Slider * )s
{
    owner = s_owner = s;
}

- (id)init
{
    if( self = [super init] )
    {
        slider = [[NSSlider alloc] initWithFrame:NSMakeRect( Slider::default_margin, 
                                                             Slider::default_margin,
                                                             Slider::default_width, 
                                                             Slider::default_slider_height )];        
        [slider setAutoresizingMask:NSViewWidthSizable | NSViewMinYMargin | NSViewMaxYMargin];
        [slider setDoubleValue:0.0];
        [slider setTarget:self];
        [slider setAction:@selector(sliderDidChange)];
        
        title = [[NSTextField alloc] initWithFrame:NSMakeRect( Slider::default_margin,
                                                               Slider::default_margin + Slider::default_slider_height + Slider::default_inner_margin,
                                                               Slider::default_width,
                                                               Slider::default_text_height )];
        [title setEditable:NO];
        [title setBezeled:NO];
        [title setDrawsBackground:NO];
        [title setAutoresizingMask:NSViewWidthSizable | NSViewMinYMargin | NSViewMaxYMargin];
        
        value = [[NSTextField alloc] initWithFrame:NSMakeRect( Slider::default_margin,
                                                               Slider::default_margin + Slider::default_slider_height + Slider::default_inner_margin,
                                                               Slider::default_width,
                                                               Slider::default_text_height )];
        [value setEditable:NO];
        [value setBezeled:NO];
        [value setDrawsBackground:NO];
        [value setAutoresizingMask:NSViewWidthSizable | NSViewMinYMargin | NSViewMaxYMargin];
        [value setAlignment:NSRightTextAlignment];
        
        t_CKFLOAT w = Slider::default_width + Slider::default_margin * 2,
            h = Slider::default_slider_height + Slider::default_inner_margin + 
            Slider::default_text_height + 
            Slider::default_margin * 2;
        master_view = [[NSView alloc] initWithFrame:NSMakeRect( 0, 0, w, h )];
        [master_view setAutoresizingMask:NSViewNotSizable];
        [master_view addSubview:slider];
        [master_view addSubview:title];
        [master_view addSubview:value];
        
        [self setDisplayFormat];
        [self updateDisplay];
    }
    
    return self;
}

- (void)destroy
{
	[title release];
	[value release];
	[slider release];
    [master_view removeFromSuperview];
    [master_view release];
    [self autorelease];
}

- (void)setName
{
    [title setStringValue:[NSString stringWithCString:owner->get_name().c_str()]];
}

- (void)setValue
{
    [slider setDoubleValue:s_owner->get_value()];
}

- (void)setRange
{
    t_CKFLOAT min = s_owner->get_min(), max = s_owner->get_max(), 
    fvalue = s_owner->get_value();
    
	[slider setMaxValue:max];
	[slider setMinValue:min];
    
    if( fvalue < min )
        s_owner->set_value( min );
    else if( fvalue > max )
        s_owner->set_value( max );
}

- (void)setDisplayFormat
{
    UI::Slider::display_format df = s_owner->get_display_format();
    
    if( df == UI::Slider::best_format )
        display_format = @"%.*g";
    else if( df == UI::Slider::integer_format )
        display_format = @"%d";
    else if( df == UI::Slider::decimal_format )
        display_format = @"%.*f";
    else if( df == UI::Slider::scientific_format )
        display_format = @"%.*e";
}

- (void)updateDisplay
{
    if( s_owner->get_display_format() == UI::Slider::integer_format )
        [value setStringValue:[NSString stringWithFormat:display_format, 
            ( t_CKINT ) lrint( [slider doubleValue] )]];
    else
        [value setStringValue:[NSString stringWithFormat:display_format, 
            s_owner->get_precision(), [slider doubleValue]]];
}

- (void)sliderDidChange
{
    s_owner->slider_changed( [slider doubleValue] );
}

@end

//------------------------------------------------------------------------------
// name: mAUIButton
// desc: ...
//------------------------------------------------------------------------------
@interface mAUIButton : mAUIElement
{
    Button * b_owner;
    NSButton * button;
}

- (void)link:( Button * )b;
- (id)init;
- (void)destroy;
- (void)setActionType;
- (void)setState;
- (void)setName;
- (void)buttonDidChange;

@end

@implementation mAUIButton

- (void)link:( Button * )b
{
    owner = b_owner = b;
}

- (id)init
{
    if( self = [super init] )
    {
        t_CKFLOAT w = Button::default_button_width + Button::default_margin * 2;
        t_CKFLOAT h = Button::default_button_height + Button::default_margin * 2;

        button = [[NSButton alloc] initWithFrame:NSMakeRect( Button::default_margin,
                                                             Button::default_margin,
                                                             Button::default_button_width,
                                                             Button::default_button_height )];
        [button setBezelStyle:NSShadowlessSquareBezelStyle];
        [button setTitle:[NSString stringWithCString:owner->get_name().c_str()]];
        [button setAutoresizingMask:NSViewWidthSizable | NSViewHeightSizable];
        
        master_view = [[NSView alloc] initWithFrame:NSMakeRect( 0, 0, w, h )];
        [master_view addSubview:button];
                        
        [button setTarget:self];
        [button setAction:@selector( buttonDidChange )];
    }
    
    return self;
}

- (void)destroy
{
    if( button != nil )
        [button release];
    if( master_view != nil )
    {
        [master_view removeFromSuperview];
        [master_view release];
    }
    
    [self autorelease];
}

- (void)setActionType
{
    Button::action_type at = b_owner->get_action_type();

    [button setState:NSOffState];
    
    if( at == Button::toggle_type )
    {
        [button setButtonType:NSPushOnPushOffButton];        
        [button sendActionOn:NSLeftMouseDownMask];
    }
    
    else if( at == Button::push_type )
    {
        [button setButtonType:NSMomentaryPushInButton];
        [button sendActionOn:NSLeftMouseDownMask | NSLeftMouseUpMask];
    }
    
}

- (void)setState
{
    if( b_owner->get_state() )
        [button setState:NSOnState];
    else
        [button setState:NSOffState];
}

- (void)setName
{
    if( button != nil )
        [button setTitle:[NSString stringWithCString:owner->get_name().c_str()]];
}

- (void)buttonDidChange
{
    b_owner->button_changed();
}

@end


//------------------------------------------------------------------------------
//  name: miniAudicleUIEButtonHelper
//  desc: ...
//------------------------------------------------------------------------------
@interface mAUIEButtonHelper : NSObject
{
    UI::Button * owner;
    NSButton * button;
}

- (id)initWithButton:(NSButton *)b
           withOwner:(UI::Button *)o;
- (void)buttonDidChange;

@end

@implementation mAUIEButtonHelper

- (id)initWithButton:(NSButton *)b
           withOwner:(UI::Button *)o
{
    if( self = [super init] )
    {
        owner = o;
        button = b;
        
        [[button cell] setAction:@selector(buttonDidChange)];
        [[button cell] setTarget:self];
    }
    
    return self;
}

- (void)dealloc
{
    [[button cell] setTarget:nil];
    [super dealloc];
}

- (void)buttonDidChange
{
    owner->button_changed();
}

@end

//namespace miniAudicle
//{
namespace UI
{

struct NativeView
{
    mAUIView * view;
};

t_CKBOOL View::init()
{
    native_data = new NativeView;

    NSAutoreleasePool * arpool = [[NSAutoreleasePool alloc] init];

    native_data->view = [mAUIView alloc];
    [native_data->view link:this];
    [native_data->view performSelectorOnMainThread:@selector(init)
                                        withObject:nil
                                     waitUntilDone:NO];
    
    [arpool release];
    
    return TRUE;
}

t_CKBOOL View::display()
{
    std::vector< Element * >::size_type i, len = elements.size();
    
    for( i = 0; i < len; i++ )
    {
        if( elements[i] )
            elements[i]->display();
    }
    
    [native_data->view performSelectorOnMainThread:@selector(display)
                                        withObject:nil
                                     waitUntilDone:NO];
        
    return TRUE;
}

t_CKBOOL View::hide()
{
    [native_data->view performSelectorOnMainThread:@selector(hide)
                                        withObject:nil
                                     waitUntilDone:NO];
    
    return TRUE;
}

t_CKBOOL View::destroy()
{
    std::vector< Element * >::size_type i, len = elements.size();
        
    delete native_data;
    
    for( i = 0; i < len; i++ )
    {
        if( elements[i] )
            elements[i]->remove_parent( this );
    }
    
    [native_data->view performSelectorOnMainThread:@selector(destroy)
                                        withObject:nil
                                     waitUntilDone:NO];
    
    return TRUE;
}

t_CKBOOL View::set_name( std::string & name )
{
    Element::set_name( name );
    
    [native_data->view performSelectorOnMainThread:@selector(setName)
                                        withObject:nil
                                     waitUntilDone:NO];
    
    return TRUE;
}

t_CKBOOL View::set_parent( View * p )
{
    return TRUE;
}

t_CKBOOL View::set_size( t_CKFLOAT w, t_CKFLOAT h )
{
    Element::set_size( w, h );
    
    [native_data->view performSelectorOnMainThread:@selector(setSize)
                                        withObject:nil
                                     waitUntilDone:NO];
    
    return TRUE;
}

t_CKBOOL View::set_position( t_CKFLOAT x, t_CKFLOAT y )
{
    Element::set_position( x, y );
    
    [native_data->view performSelectorOnMainThread:@selector(setPosition)
                                        withObject:nil
                                     waitUntilDone:NO];
    
    return TRUE;
}

t_CKBOOL View::add_element( Element * e )
{
    if( e == NULL )
        return FALSE;
    
    elements.push_back( e );
    e->set_parent( this );
    
    return TRUE;
}

t_CKBOOL View::remove_element( Element * e )
{
    if( e == NULL )
        return FALSE;
    
    std::vector< Element * >::size_type i, len = elements.size();
    
    for( i = 0; i < len; i++ )
    {
        if( elements[i] == e )
        {
            elements[i]->remove_parent( this );
            elements[i] = NULL;
        }
    }
    
    return TRUE;
}

//------------------------------------------------------------------------------
//  name: NativeSlider
//  desc: native data for slider
//------------------------------------------------------------------------------
struct NativeSlider
{
    mAUISlider * slider;
};

const t_CKUINT Slider::default_margin = 18;
const t_CKUINT Slider::default_width = 202;
const t_CKUINT Slider::default_slider_height = 21;
const t_CKUINT Slider::default_inner_margin = 5;
const t_CKUINT Slider::default_text_height = 17;

t_CKBOOL Slider::init()
{	
    native_data = new NativeSlider;
    
    w = default_width + default_margin * 2;
    h = default_slider_height + default_inner_margin + 
        default_text_height + default_margin * 2;
    
	name = "";
	e = NULL;
    
    df = best_format;
    precision = 4;
    	
    native_data->slider = [mAUISlider alloc];
    [native_data->slider link:this];
    [native_data->slider performSelectorOnMainThread:@selector(init)
                                          withObject:nil
                                       waitUntilDone:NO];
	
	return TRUE;
}

t_CKBOOL Slider::display()
{
    Element::display();
    
    [native_data->slider performSelectorOnMainThread:@selector(display)
                                          withObject:nil
                                       waitUntilDone:NO];
    
	return TRUE;
}

t_CKBOOL Slider::hide()
{
    Element::hide();
    
    [native_data->slider performSelectorOnMainThread:@selector(hide)
                                          withObject:nil
                                       waitUntilDone:NO];
        
    return TRUE;
}

t_CKBOOL Slider::destroy()
{
    Element::destroy();
    
    [native_data->slider performSelectorOnMainThread:@selector(destroy)
                                          withObject:nil
                                       waitUntilDone:NO];
    delete native_data;
        
	return TRUE;
}

t_CKBOOL Slider::set_name( string & name )
{
    Element::set_name( name );
	
    [native_data->slider performSelectorOnMainThread:@selector(setName)
                                          withObject:nil
                                       waitUntilDone:NO];
    
	return TRUE;
}

t_CKBOOL Slider::set_parent( View * p )
{
    Element::set_parent( p );
    
    [parent->native_data->view performSelectorOnMainThread:@selector(addElement:)
                                                withObject:native_data->slider
                                             waitUntilDone:NO];
    
    return TRUE;
}

t_CKBOOL Slider::remove_parent( View * p )
{
    Element::remove_parent( p );
    
    [native_data->slider performSelectorOnMainThread:@selector(removeParent)
                                          withObject:nil
                                       waitUntilDone:NO];
    
    return TRUE;
}

t_CKBOOL Slider::set_size( t_CKFLOAT w, t_CKFLOAT h )
{
    Element::set_size( w, h );

    [native_data->slider performSelectorOnMainThread:@selector(setSize)
                                          withObject:nil
                                       waitUntilDone:NO];
        
    return TRUE;
}

t_CKBOOL Slider::set_position( t_CKFLOAT x, t_CKFLOAT y )
{
    Element::set_position( x, y );
    
    [native_data->slider performSelectorOnMainThread:@selector(setPosition)
                                          withObject:nil
                                       waitUntilDone:NO];
    
    return TRUE;
}

t_CKDOUBLE Slider::get_value()
{	
	return value;
}

t_CKBOOL Slider::set_value( t_CKDOUBLE v )
{
    value = v;
    
    if( value > max )
        value = max;
    if( value < min )
        value = min;
    
    [native_data->slider performSelectorOnMainThread:@selector(setValue)
                                          withObject:nil
                                       waitUntilDone:NO];
    
    [native_data->slider performSelectorOnMainThread:@selector(updateDisplay)
                                          withObject:nil
                                       waitUntilDone:NO];
    
	return TRUE;
}

t_CKBOOL Slider::set_range( t_CKDOUBLE low, t_CKDOUBLE high )
{
    min = low;
    max = high;
    [native_data->slider performSelectorOnMainThread:@selector(setRange)
                                          withObject:nil
                                       waitUntilDone:NO];
        
    set_value( get_value() );
		
	return TRUE;
}

t_CKDOUBLE Slider::get_max()
{
	return max;
}

t_CKDOUBLE Slider::get_min()
{
	return min;
}

t_CKBOOL Slider::set_precision( t_CKUINT precision )
{
    this->precision = precision;
    
    // refresh
    set_value( get_value() );
    
    return TRUE;
}

t_CKUINT Slider::get_precision()
{
    return precision;
}

t_CKBOOL Slider::set_display_format( Slider::display_format f )
{
    this->df = f;
        
    [native_data->slider performSelectorOnMainThread:@selector(setDisplayFormat)
                                          withObject:nil
                                       waitUntilDone:NO];
    
    // refresh
    set_value( get_value() );
    
    return TRUE;
}

Slider::display_format Slider::get_display_format()
{
    return df;
}

void Slider::slider_changed( t_CKDOUBLE v )
{
	if( df == integer_format )
        value = ( t_CKINT ) lrint( v );
    else
        value = v;
    
    if( e != nil )
		e->queue_broadcast();
    
    // refresh
    [native_data->slider performSelectorOnMainThread:@selector(updateDisplay)
                                          withObject:nil
                                       waitUntilDone:NO];
}

struct NativeButton
{
    mAUIButton * button;
};

const t_CKUINT Button::default_margin = 18;
const t_CKUINT Button::default_button_width = 55;
const t_CKUINT Button::default_button_height = 55;

t_CKBOOL Button::init()
{
    x = y = 0;
    w = Button::default_button_width + Button::default_margin * 2;
    h = Button::default_button_height + Button::default_margin * 2;
    name = "";
    
    pushed = FALSE;
    at = push_type;
    
    native_data = new NativeButton;
    native_data->button = [mAUIButton alloc];
    [native_data->button link:this];
    [native_data->button performSelectorOnMainThread:@selector( init )
                                           withObject:nil
                                        waitUntilDone:NO];
        
    return TRUE;
}

t_CKBOOL Button::display()
{
    Element::display();
    
    [native_data->button performSelectorOnMainThread:@selector( display )
                                          withObject:nil
                                       waitUntilDone:NO];
        
    return TRUE;
}

t_CKBOOL Button::hide()
{
    Element::hide();
    
    [native_data->button performSelectorOnMainThread:@selector( hide )
                                          withObject:nil
                                       waitUntilDone:NO];
    
    return TRUE;
}

t_CKBOOL Button::destroy()
{
    Element::destroy();
        
    [native_data->button performSelectorOnMainThread:@selector( destroy )
                                          withObject:nil
                                       waitUntilDone:NO];
    delete native_data;
        
    return TRUE;
}

t_CKBOOL Button::set_name( std::string & name )
{
    Element::set_name( name );

    [native_data->button performSelectorOnMainThread:@selector( setName )
                                          withObject:nil
                                       waitUntilDone:NO];
    
    return TRUE;
}

t_CKBOOL Button::set_size( t_CKFLOAT w, t_CKFLOAT h )
{
    Element::set_size( w, h );
    
    [native_data->button performSelectorOnMainThread:@selector( setSize )
                                          withObject:nil
                                       waitUntilDone:NO];
    
    return TRUE;
}

t_CKBOOL Button::set_position( t_CKFLOAT x, t_CKFLOAT y )
{
    Element::set_position( x, y );
    
    [native_data->button performSelectorOnMainThread:@selector( setPosition )
                                          withObject:nil
                                       waitUntilDone:NO];
    
    return TRUE;
}

t_CKBOOL Button::set_parent( View * p )
{
    Element::set_parent( p );
    
    [parent->native_data->view performSelectorOnMainThread:@selector(addElement:)
                                                withObject:native_data->button
                                             waitUntilDone:NO];
    
    return TRUE;
}

t_CKBOOL Button::remove_parent( View * p )
{
    Element::remove_parent( p );
    
    [native_data->button performSelectorOnMainThread:@selector(removeParent)
                                          withObject:nil
                                       waitUntilDone:NO];
    
    return TRUE;
}

t_CKBOOL Button::get_state()
{
    return pushed;
}

t_CKBOOL Button::set_state( t_CKBOOL state )
{
    if( state )
        pushed = 1;
    else
        pushed = 0;
    
    [native_data->button performSelectorOnMainThread:@selector(setState)
                                          withObject:nil
                                       waitUntilDone:NO];
    
    return FALSE;
}

t_CKBOOL Button::set_action_type( action_type t )
{
    if( t == toggle_type || t == push_type )
    {
        if( pushed )
            button_changed();
        at = t;
        [native_data->button performSelectorOnMainThread:@selector(setActionType)
                                              withObject:nil
                                           waitUntilDone:NO];
        return TRUE;
    }
    
    return FALSE;
}

Button::action_type Button::get_action_type()
{
    return at;
}

void Button::button_changed()
{
    if( e != NULL )
        e->queue_broadcast();
    
    pushed = !pushed;
}

struct NativeLED
{
    NSImage * off_image;
    NSImage * on_image;
    NSImageView * image_view;
    NSView * master_view;
    
    static const NSString * off_image_name;
    static const NSString * red_image_name;
    static const NSString * green_image_name;
    static const NSString * blue_image_name;
    static const t_CKUINT margin = 18;
    static const t_CKFLOAT default_flash_time;
};

const NSString * NativeLED::off_image_name = @"button-off.png";
const NSString * NativeLED::red_image_name = @"button-red.png";
const NSString * NativeLED::green_image_name = @"button-green.png";
const NSString * NativeLED::blue_image_name = @"led-blue.png";
const t_CKFLOAT NativeLED::default_flash_time = .150;

t_CKBOOL LED::init()
{
    
    NSAutoreleasePool * arpool = [[NSAutoreleasePool alloc] init];
    
    native_data = new NativeLED;
    
    native_data->on_image = [[NSImage imageNamed:native_data->red_image_name] retain];
    native_data->off_image = [[NSImage imageNamed:native_data->off_image_name] retain];
    c = red;
    
    native_data->image_view = [[NSImageView alloc] initWithFrame:NSMakeRect( NativeLED::margin,
                                                                             NativeLED::margin,
                                                                             28, 28 )];
    [native_data->image_view setImage:native_data->off_image];
    [native_data->image_view setEnabled:NO];
    [native_data->image_view setAutoresizingMask:
        NSViewMinXMargin | NSViewMaxXMargin | NSViewMinYMargin | NSViewMaxYMargin];
    
    NSRect r = [native_data->image_view frame];
    r.origin.x = 0;
    r.origin.y = 0;
    r.size.height += NativeLED::margin * 2;
    r.size.width += NativeLED::margin * 2;
    native_data->master_view = [[NSView alloc] initWithFrame:r];
    
    [native_data->master_view addSubview:native_data->image_view];
    
    x = y = 0;
    w = r.size.width;
    h = r.size.height;
    
    flash_time = NativeLED::default_flash_time;
    
    [arpool release];
    
    return TRUE;
}

t_CKBOOL LED::display()
{
    Element::display();
    
    NSAutoreleasePool * arpool = [[NSAutoreleasePool alloc] init];
    
    [native_data->master_view setHidden:NO];
    
    [arpool release];
    
    return TRUE;
}

t_CKBOOL LED::hide()
{
    Element::hide();
    
    NSAutoreleasePool * arpool = [[NSAutoreleasePool alloc] init];

    [native_data->master_view setHidden:YES];
    
    [arpool release];
    
    return TRUE;
}

t_CKBOOL LED::destroy()
{
    Element::destroy();
    
    NSAutoreleasePool * arpool = [[NSAutoreleasePool alloc] init];
    
    [native_data->image_view release];
    [native_data->on_image release];
    [native_data->off_image release];
    [native_data->master_view removeFromSuperview];
    [native_data->master_view release];
    delete native_data;
    
    [arpool release];
    
    return TRUE;
}

t_CKBOOL LED::set_size( t_CKFLOAT nw, t_CKFLOAT nh )
{
    Element::set_size( nw, nh );
    
    NSAutoreleasePool * arpool = [[NSAutoreleasePool alloc] init];
    
    [native_data->master_view setFrameSize:NSMakeSize( nw, nh )];
    
    [arpool release];
    
    return TRUE;
}

t_CKBOOL LED::set_position( t_CKFLOAT nx, t_CKFLOAT ny )
{
    Element::set_position( nx, ny );
    
    NSAutoreleasePool * arpool = [[NSAutoreleasePool alloc] init];
    
    [native_data->master_view setFrameOrigin:NSMakePoint( nx, ny )];
    
    [arpool release];
    
    return TRUE;
}

t_CKBOOL LED::set_parent( View * p )
{
    Element::set_parent( p );

    NSAutoreleasePool * arpool = [[NSAutoreleasePool alloc] init];
    
    [parent->native_data->view performSelectorOnMainThread:@selector(addSubview:)
                                                withObject:native_data->master_view
                                             waitUntilDone:NO];
        
    [arpool release];
    
    return TRUE;
}

t_CKBOOL LED::remove_parent( View * p )
{
    Element::remove_parent( p );
    
    NSAutoreleasePool * arpool = [[NSAutoreleasePool alloc] init];
    
    [native_data->master_view removeFromSuperview];
    
    [arpool release];
    
    return TRUE;
}

t_CKBOOL LED::light()
{
    NSAutoreleasePool * arpool = [[NSAutoreleasePool alloc] init];
    
    [native_data->image_view performSelectorOnMainThread:@selector(setImage:)
                                              withObject:native_data->on_image
                                           waitUntilDone:NO];
    
    [arpool release];

    return TRUE;
}

t_CKBOOL LED::unlight()
{
    NSAutoreleasePool * arpool = [[NSAutoreleasePool alloc] init];
    
    [native_data->image_view performSelectorOnMainThread:@selector(setImage:)
                                              withObject:native_data->off_image
                                           waitUntilDone:NO];
     
    [arpool release];

    return TRUE;
}

t_CKBOOL LED::set_color( color c )
{
    NSAutoreleasePool * arpool = [[NSAutoreleasePool alloc] init];
    
    unlight();
    
    if( c == red )
    {
        [native_data->on_image release];
        native_data->on_image = [[NSImage imageNamed:native_data->red_image_name] retain];
        this->c = red;
    }
    else if( c == green )
    {
        [native_data->on_image release];
        native_data->on_image = [[NSImage imageNamed:native_data->green_image_name] retain];
        this->c = green;
    }
    
    else if( c == blue )
    {
        [native_data->on_image release];
        native_data->on_image = [[NSImage imageNamed:native_data->blue_image_name] retain];
        this->c = blue;
    }
    
    [arpool release];

    return TRUE;
}

LED::color LED::get_color()
{
    return c;
}

t_CKBOOL Mouse::init()
{
    active = FALSE;
    return TRUE;
}

t_CKBOOL Mouse::display()
{
    active = TRUE;
    
    return TRUE;
}

t_CKBOOL Mouse::hide()
{
    active = FALSE;
    
    return TRUE;
}

t_CKBOOL Mouse::destroy()
{
    // nop
    
    return TRUE;
}

t_CKUINT Mouse::x()
{
    NSPoint p = [NSEvent mouseLocation];
    return (t_CKUINT)p.x;
}

t_CKUINT Mouse::y()
{
    NSPoint p = [NSEvent mouseLocation];
    return (t_CKUINT)p.y;
}

t_CKUINT Mouse::maxx()
{
    NSRect p = [[NSScreen mainScreen] visibleFrame];
    return (t_CKUINT)p.size.width;
}

t_CKUINT Mouse::maxy()
{
    NSRect p = [[NSScreen mainScreen] visibleFrame];
    return (t_CKUINT)p.size.height;
}

t_CKDOUBLE Mouse::fx()
{
    return ((t_CKDOUBLE)x())/((t_CKDOUBLE)maxx());
}

t_CKDOUBLE Mouse::fy()
{
    return ((t_CKDOUBLE)y())/((t_CKDOUBLE)maxy());
}


} /* UI */
//} /* miniAudicle */
