/*----------------------------------------------------------------------------
miniAudicle
Cocoa GUI to chuck audio programming environment

Copyright (c) 2005 Spencer Salazar.  All rights reserved.
http://chuck.cs.princeton.edu/
http://soundlab.cs.princeton.edu/

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
U.S.A.
-----------------------------------------------------------------------------*/

//-----------------------------------------------------------------------------
// file: miniaudicle.cpp
// desc: Platform independent miniAudicle interface
//
// author: Spencer Salazar (ssalazar@princeton.edu)
// date: Autumn 2005
//-----------------------------------------------------------------------------
#include <stdio.h>
#include <errno.h>
#include <time.h>
#include <string.h>
#include <ctype.h>

#ifdef __PLATFORM_WIN32__
#include <windows.h>
#include <io.h>
#include <fcntl.h>
#else
#include <unistd.h>
#endif

#include "miniAudicle.h"
#include "digiio_rtaudio.h"
#include "chuck_otf.h"
#include "chuck_errmsg.h"
#include "chuck_globals.h"

//#ifdef __MACOSX_CORE__
#include "miniAudicle_ui_elements.h"
#include "miniAudicle_import.h"
//using namespace miniAudicle;
//#endif

// default destination host name
// extern char g_host[256];

t_CKBOOL g_forked = FALSE;

#if defined(__MACOSX_CORE__)
t_CKINT priority = 80;
t_CKINT priority_low = 60;
#elif defined(__PLATFORM_WIN32__)
t_CKINT priority = 0;
t_CKINT priority_low = 0;
#else
t_CKINT priority = 0x7fffffff;
t_CKINT priority_low = 0x7fffffff;
#endif

extern const char MA_VERSION[] = "0.1.3.6c (alpha)\0";
extern const char MA_ABOUT[] = "version %s\nCopyright (c) Spencer Salazar\n\nChucK: version %s\nCopyright (c) Ge Wang and Perry Cook\nhttp://chuck.cs.princeton.edu/\0";
extern const char MA_HELP[] = 
"usage: miniAudicle [options] [files] \n\
options: \n\
 --dacN           use audio output device N\n\
 --adcN           use audio input device N\n\
 --outN/-oN       use N output channels\n\
 --inN/-iN        use N input channels\n\
 --channelsN/-cN  use N input/output channels\n\
 --srateN/-sN     set sample rate to N\n\
 --bufsizeN/-bN   set sample buffer size to N (rounded to nearest power of 2)\n\
 --bufnumN/-nN    use N sample buffers\n\
 --verboseN/-vN   set log level to N (0-10:none-everything)\n\
 --probe          list available audio devices and properties\n\
 --help/--about   print this message\n\
\n\
miniAudicle-%s\n\
http://audicle.cs.princeton.edu/mini/\n\
\n\
ChucK-%s\n\
http://chuck.cs.princeton.edu/\n";

//-----------------------------------------------------------------------------
// name: vm_cb
// desc: thread routine to run vm 
//-----------------------------------------------------------------------------
void * vm_cb( void * v )
{	
	// boost priority
	if( Chuck_VM::our_priority != 0x7fffffff )
	{
		// try
		if( !Chuck_VM::set_priority( Chuck_VM::our_priority, g_vm ) )
		{
			// error
			fprintf( stderr, "[chuck]: %s\n", g_vm->last_error() );
			return FALSE;
		}
	}
	
	
	// run the vm
    g_vm->run();
	
    // detach
	all_detach();
    
    // log
    EM_log( CK_LOG_SEVERE, "VM callback process ending..." );
	
    // free vm
    //g_vm = NULL; SAFE_DELETE( g_vm );
	//SAFE_DELETE( g_vm );
    // free the compiler
    //SAFE_DELETE( compiler );
	
	return NULL;
}

//-----------------------------------------------------------------------------
// name: miniAudicle()
// desc: ... 
//-----------------------------------------------------------------------------
miniAudicle::miniAudicle()
{
	vm = NULL;
	vm_on = FALSE;
    
    class_names = new map< string, t_CKINT >;
    
    next_document_id = 0;

    vm_options.enable_audio = TRUE;
	vm_options.enable_network = FALSE;
    vm_options.dac = 0;
    vm_options.adc = 0;
    vm_options.srate = SAMPLING_RATE_DEFAULT;
    vm_options.buffer_size = BUFFER_SIZE_DEFAULT;
    vm_options.num_buffers = NUM_BUFFERS_DEFAULT;
	vm_options.num_inputs = 2;
	vm_options.num_outputs = 2;
    
    probe();    
}

//-----------------------------------------------------------------------------
// name: ~miniAudicle()
// desc: ...
//-----------------------------------------------------------------------------
miniAudicle::~miniAudicle()
{
	if( vm_on )
		stop_vm();

    delete class_names;
    
	// log
    EM_log( CK_LOG_INFO, "miniAudicle instance destroyed..." );
}

//-----------------------------------------------------------------------------
// name: run_code()
// desc: ...
//-----------------------------------------------------------------------------
t_CKBOOL miniAudicle::run_code( string & code, string & name, t_CKUINT docid, 
                                t_CKUINT & shred_id, string & out )
{
	FILE * read, * write;
    
    if( documents.find( docid ) == documents.end() )
        // invalid document id
    {
        out += "internal error at miniAudicle::run_code\n";
        return FALSE;
    }

    // write code to FILE *-accessible buffer
#ifndef __PLATFORM_WIN32__
    read = tmpfile();
	if( read == NULL )
	{
		EM_log( CK_LOG_SEVERE, "(miniAudicle): error in tmpfile: %d", errno );
        return FALSE;
	}
    write = read;

#else
	HANDLE hRead, hWrite;
	int fd_read, fd_write;

	if( !CreatePipe( &hRead, &hWrite, NULL, code.size() ) )
	{
		EM_log( CK_LOG_SEVERE, "(miniAudicle): error in CreatePipe: %d", GetLastError() );
		return FALSE;
	}

	// fails on 64 bit Windows!
	fd_read = _open_osfhandle( ( long ) hRead, _O_RDONLY );
	fd_write = _open_osfhandle( ( long ) hWrite, _O_WRONLY );

	read = _fdopen( fd_read, "r" );
	write = _fdopen( fd_write, "w" );

#endif
    
	// write the code to the pipe
	fprintf( write, "%s", code.c_str() );

	if( read != write )
		fclose( write );
	
	// compile
	if( !compiler->go( name, read ) )
	{
		out += string( EM_lasterror() ) + "\n";
        fclose( read ); read = NULL;
		return false;
	}
    fclose( read ); read = NULL;
	
	// allocate the VM message struct
	Chuck_Msg * msg = new Chuck_Msg;

	// fill in the VM message
	msg->code = compiler->output();  //do i need to fclose( read )?
		
	msg->code->name = name;
	msg->type = MSG_ADD;
	msg->reply = ( ck_msg_func )1;
		
	// execute
	vm->queue_msg( msg, 1 );
	
	// check results
	while( ( msg = vm->get_reply() ) == NULL )
		usleep( 10000 );
	
    shred_id = msg->replyA;
	if( shred_id == 0 )
	{
		out += string( EM_lasterror() ) + "\n";
		delete msg;
		return FALSE;
	}
	
    documents[docid]->push_back( shred_id );
    shreds[shred_id].docid = docid;
    shreds[shred_id].index = documents[docid]->size() - 1;
    
	delete msg;
    
	return TRUE;
}

//-----------------------------------------------------------------------------
// name: replace_code()
// desc: ...
//-----------------------------------------------------------------------------
t_CKBOOL miniAudicle::replace_code( string & code, string & name, t_CKUINT docid, 
                                    t_CKUINT & shred_id, string & out )
{	 
	if( documents.find( docid ) == documents.end() )
	{
        out += "internal error at miniAudicle::replace_code\n";
		return FALSE;
	}
    
    while( documents[docid]->size() && documents[docid]->back() == 0 )
        documents[docid]->pop_back();
    
    if( documents[docid]->size() == 0 )
    {
        out += "no shred to replace\n";
        return FALSE;
    }

    // create a FILE *-accessible buffer
	FILE * read, * write;
#ifndef __PLATFORM_WIN32__
    read = tmpfile();
	if( read == NULL )
	{
		EM_log( CK_LOG_SEVERE, "(miniAudicle): error in tmpfile: %d", errno );
        return FALSE;
	}
    write = read;

#else
	HANDLE hRead, hWrite;
	int fd_read, fd_write;

	if( !CreatePipe( &hRead, &hWrite, NULL, code.size() ) )
	{
		EM_log( CK_LOG_SEVERE, "(miniAudicle): error in CreatePipe: %d", GetLastError() );
		return FALSE;
	}

	// WARNING: fails on 64 bit Windows!
	fd_read = _open_osfhandle( ( long ) hRead, _O_RDONLY );
	fd_write = _open_osfhandle( ( long ) hWrite, _O_WRONLY );

	read = _fdopen( fd_read, "r" );
	write = _fdopen( fd_write, "w" );

#endif
    
    fprintf( write, "%s", code.c_str() );
	if( read != write )
		fclose( write );
	
	if( !compiler->go( name, read ) )
	{
		out += string( EM_lasterror() ) + "\n";
        fclose( read ); read = NULL;
		return false;
	}
    fclose( read ); read = NULL;
	
	Chuck_Msg * msg = new Chuck_Msg;

	msg->code = compiler->output();
	
	msg->code->name = name;
	msg->type = MSG_REPLACE;
	msg->param = documents[docid]->back();
	msg->reply = ( ck_msg_func )1;
	
	vm->queue_msg( msg, 1 );
	
	// check results
	while( ( msg = vm->get_reply() ) == NULL )
		usleep( 10000 );
	
    shred_id = msg->replyA;
	if( shred_id == 0 )
	{
		out += string( EM_lasterror() ) + "\n";
        shreds.erase( documents[docid]->back() );
        documents[docid]->pop_back(); // shred probably no longer exists
		delete msg;
		return FALSE;
	}
	
    // in case the shred id changed, recreate the various map entries
    shreds.erase( documents[docid]->back() );
    documents[docid]->back() = shred_id;
    shreds[shred_id].docid = docid;
    shreds[shred_id].index = documents[docid]->size() - 1;
    
	delete msg;
	
	return TRUE;
}

//-----------------------------------------------------------------------------
// name: remove_code()
// desc: ...
//-----------------------------------------------------------------------------
t_CKBOOL miniAudicle::remove_code( t_CKUINT docid, t_CKUINT & shred_id, 
                                   string & out )
{
	if( documents.find( docid ) == documents.end() )
	{
        out += "internal error at miniAudicle::remove_code\n";
		return FALSE;
	}
    
    while( documents[docid]->size() && documents[docid]->back() == 0 )
        documents[docid]->pop_back();
    
    if( documents[docid]->size() == 0 )
    {
        out += "no shred to remove\n";
        return FALSE;
    }
    
	Chuck_Msg * msg = new Chuck_Msg;
	
	msg->type = MSG_REMOVE;
	msg->param = documents[docid]->back();
	msg->reply = ( ck_msg_func )1;
	
	vm->queue_msg( msg, 1 );
	
	// check results
	while( ( msg = vm->get_reply() ) == NULL )
		usleep( 10000 );
	
	if( msg->replyA == 0 )
	{
		out += string( EM_lasterror() ) + "\n";
        shreds.erase( documents[docid]->back() );
        documents[docid]->pop_back(); // shred probably no longer exists
        delete msg;
 		return FALSE;
	}
	
    shred_id = documents[docid]->back();
    shreds.erase( shred_id );
    documents[docid]->pop_back();
    
	delete msg;
	
	return TRUE;
}

//-----------------------------------------------------------------------------
// name: remove_shred()
// desc: ...
//-----------------------------------------------------------------------------
t_CKBOOL miniAudicle::remove_shred( t_CKINT shred_id, string & out )
{
	Chuck_Msg * msg = new Chuck_Msg;
    
	msg->type = MSG_REMOVE;
	msg->param = shred_id;
	msg->reply = ( ck_msg_func )1;
	
	vm->queue_msg( msg, 1 );
	
	// check results
	while( ( msg = vm->get_reply() ) == NULL )
		usleep( 10000 );
	
	if( msg->replyA == 0 )
	{
		out += string( EM_lasterror() ) + "\n";
		return FALSE;
	}
	
    // set the corresponding document-shred association to 0, so future calls
    // to document-based otf functions will ignore that association
    if( shreds.count( shred_id ) ) // make sure the entry exists first...
    {
        documents[shreds[shred_id].docid]->at( shreds[shred_id].index ) = 0;
        shreds.erase( shred_id );
    }
    
	delete msg;
	
	return TRUE;
}

//-----------------------------------------------------------------------------
// name: removeall()
// desc: ...
//-----------------------------------------------------------------------------
t_CKBOOL miniAudicle::removeall( string & out )
{
	Chuck_Msg * msg = new Chuck_Msg;
	
	msg->type = MSG_REMOVEALL;
	msg->reply = ( ck_msg_func )1;
	
	vm->queue_msg( msg, 1 );
	
	// check results
	while( ( msg = vm->get_reply() ) == NULL )
		usleep( 10000 );
	
	if( msg->replyA == 0 )
	{
		out += string( EM_lasterror() ) + "\n";
		return FALSE;
	}
	
	delete msg;
	
	return TRUE;
}

//-----------------------------------------------------------------------------
// name: removelast()
// desc: ...
//-----------------------------------------------------------------------------
t_CKBOOL miniAudicle::removelast( string & out )
{
	Chuck_Msg * msg = new Chuck_Msg;
	
	msg->type = MSG_REMOVE;
	msg->param = 0xffffffff;
	msg->reply = ( ck_msg_func )1;
	
	vm->queue_msg( msg, 1 );
	
	// check results
	while( ( msg = vm->get_reply() ) == NULL )
		usleep( 10000 );
	
	if( msg->replyA == 0 )
	{
		out += string( EM_lasterror() ) + "\n";
		return FALSE;
	}
	
	delete msg;
	
	return TRUE;
}

t_CKUINT miniAudicle::allocate_document_id()
{    
    next_document_id++;
    
    // ensure that we have a next_document_id that isnt actually in use
    // if there are more than INT_MAX documents this will infinite loop
    while( documents.find( next_document_id ) != documents.end() )
        next_document_id++;
    
    documents[next_document_id] = new vector< t_CKUINT >;
    
    return next_document_id;
}

void miniAudicle::free_document_id( t_CKUINT docid )
{
    if( documents.find( docid ) != documents.end() )
    {
        // delete data
        delete documents[docid];
        // remove from map
        documents.erase( docid );
    }
}

//-----------------------------------------------------------------------------
// name: get_log_level()
// desc: ...
//-----------------------------------------------------------------------------
int miniAudicle::get_log_level()
{
	return g_loglevel;
}

//-----------------------------------------------------------------------------
// name: set_log_level()
// desc: ...
//-----------------------------------------------------------------------------
t_CKBOOL miniAudicle::set_log_level( int l )
{
	EM_setlog( l );
	return FALSE;
}

//-----------------------------------------------------------------------------
// name: start_vm()
// desc: ...
//-----------------------------------------------------------------------------
t_CKBOOL miniAudicle::start_vm()
{
    char buffer[1024];
    time_t t;
    
    // clear the class name existence map
    class_names->clear();
    
    time(&t);
    strncpy( buffer, ctime(&t), 24 );
    buffer[24] = '\0';

    // log
    EM_log( CK_LOG_SYSTEM, "-------( %s )-------", buffer );
    EM_log( CK_LOG_SYSTEM, "starting chuck virtual machine..." );
    // push log
    EM_pushlog();
    
    if( vm == NULL )
    {
        // log
        EM_log( CK_LOG_INFO, "allocating VM..." );
        t_CKBOOL enable_audio = vm_options.enable_audio;
        t_CKBOOL vm_halt = FALSE;
        t_CKUINT srate = vm_options.srate;
        t_CKUINT buffer_size = vm_options.buffer_size;
        t_CKUINT num_buffers = NUM_BUFFERS_DEFAULT;
        t_CKUINT dac = vm_options.dac;
        t_CKUINT adc = vm_options.adc;
        t_CKBOOL set_priority = FALSE;
        t_CKBOOL block = FALSE;
        t_CKUINT output_channels = vm_options.num_outputs;
        t_CKUINT input_channels = vm_options.num_inputs;

        // check buffer size
        // buffer_size = next_power_2( buffer_size-1 );
        // audio, boost
        if( !set_priority && !block ) priority = priority_low;
        if( !set_priority && !enable_audio ) priority = 0x7fffffff;
        // set priority
        Chuck_VM::our_priority = priority;
        // set watchdog
#ifdef __MACOSX_CORE__
        g_do_watchdog = TRUE;
        g_watchdog_timeout = .5;
#else
        g_do_watchdog = FALSE;
#endif
        
        // allocate the vm - needs the type system
        vm = g_vm = new Chuck_VM;

        if( !vm->initialize( enable_audio, vm_halt, srate, buffer_size,
                             num_buffers, dac, adc, output_channels, 
                             input_channels, block ) )
        {
            fprintf( stderr, "[chuck]: %s\n", vm->last_error() );
            // pop
            EM_poplog();
            return FALSE;
        }
            
        // log
        EM_log( CK_LOG_INFO, "allocating compiler..." );

        // allocate the compiler
        g_compiler = compiler = new Chuck_Compiler;
        // initialize the compiler
        compiler->initialize( vm );
        // enable dump
        compiler->emitter->dump = FALSE;
        // set auto depend
        compiler->set_auto_depend( FALSE );

        // vm synthesis subsystem - needs the type system
        if( !vm->initialize_synthesis() )
        {
            fprintf( stderr, "[chuck]: %s\n", vm->last_error() );
            // pop
            EM_poplog();
            return FALSE;
        }
        
#ifdef __MACOSX_CORE__
        // import api
        init_maui( compiler->env );
#endif

        // reset the parser
        reset_parse();
        
		// start the vm handler threads
#ifndef __PLATFORM_WIN32__
		pthread_create( &vm_tid, NULL, vm_cb, NULL );
#else
		vm_tid = CreateThread( NULL, 0, (LPTHREAD_START_ROUTINE)vm_cb, NULL, 0, 0 );
#endif
    }

    // check it
	if( !g_forked && vm_options.enable_network )
	{
		// start tcp server
		g_sock = ck_tcp_create( 1 );
		if( !g_sock || !ck_bind( g_sock, g_port ) || !ck_listen( g_sock, 10 ) )
		{
			fprintf( stderr, "[chuck]: cannot bind to tcp port %i...\n", g_port );
			ck_close( g_sock );
			g_sock = NULL;
		}
		else
		{
#ifndef __PLATFORM_WIN32__
			pthread_create( &otf_tid, NULL, otf_cb, NULL );
#else
			otf_tid = CreateThread( NULL, 0, (LPTHREAD_START_ROUTINE)otf_cb, NULL, 0, 0 );
#endif
		}
        
        g_forked = TRUE;
	}
    
    vm_on = TRUE;
    // pop
    EM_poplog();
    
	return vm_on;
}

//-----------------------------------------------------------------------------
// name: stop_vm()
// desc: ...
//-----------------------------------------------------------------------------
t_CKBOOL miniAudicle::stop_vm()
{
	// if it's there
    if( vm )
    {
        EM_log( CK_LOG_SYSTEM, "stopping chuck virtual machine..." );
        // get vm
        Chuck_VM * the_vm = g_vm;
        // flag the global one
        vm = g_vm = NULL;
        // if not NULL
        if( the_vm && vm_on )
        {
			// flag
			vm_on = FALSE;

            // stop
            the_vm->stop();
        
            // wait a bit
            usleep( 100000 );

            // detach
            // all_detach();

#if !defined(__PLATFORM_WIN32__) && !defined(__WINDOWS_PTHREAD__)
            SAFE_DELETE( the_vm );
#endif
        }

        SAFE_DELETE( compiler );
        g_compiler = compiler = NULL;
    }

	return TRUE;
}

//-----------------------------------------------------------------------------
// name: is_on()
// desc: ...
//-----------------------------------------------------------------------------
t_CKBOOL miniAudicle::is_on()
{
	return vm_on;
}

//-----------------------------------------------------------------------------
// name: shred_count()
// desc: ...
//-----------------------------------------------------------------------------
t_CKUINT miniAudicle::shred_count()
{
	if( vm_on != TRUE || vm == NULL )
		return 0;
	Chuck_VM_Status status;
    vm->shreduler()->status( &status );
	return (t_CKUINT)status.list.size();
}

//-----------------------------------------------------------------------------
// name: status()
// desc: ...
//-----------------------------------------------------------------------------
t_CKINT miniAudicle::status( Chuck_VM_Status * status)
{
	if( vm_on != TRUE || vm == NULL )
		return 0;
	vm->shreduler()->status( status );
	return 1;
}

void tokenize_string( string & str, vector< string > & tokens)
{
    t_CKINT space = 1;
    t_CKINT end_space = 0;
    t_CKINT dquote = 0;
    t_CKINT i = 0, j = 0, len = str.size();
    
    for( i = 0; i < len; i++ )
    {
        if( isspace( str[i] ) && space )
        {
            j++;
            continue;
        }
        
        if( isspace( str[i] ) && end_space )
        {
            tokens.push_back( string( str, j, i - j ) );
            j = i + 1;
            space = 1;
            end_space = 0;
            continue;
        }
        
        if( str[i] == '"' ) 
        {
            if( !dquote )
            {
                str.erase( i, 1 );
                i--;
                len--;
                space = 0;
                end_space = 0;
                dquote = 1;
                continue;
            }
            
            else if( str[i - 1] == '\\' )
            {
                str.erase( i - 1, 1 );
                len--;
                i--;
                continue;
            }
            
            else
            {
                str.erase( i, 1 );
                i--;
                len--;
                dquote = 0;
                end_space = 1;
                space = 0;
                continue;
            }
        }

        if( !dquote )
        {       
            end_space = 1;
            space = 0;
        }
    }

    if( i > j && end_space )
    {
        tokens.push_back( string( str, j, i - j ) );
    }
}

//-----------------------------------------------------------------------------
// name: highlight_line()
// desc: ...
//-----------------------------------------------------------------------------
t_CKBOOL miniAudicle::highlight_line( string & line, 
                                      miniAudicle_SyntaxHighlighting * sh )
{
    vector< string > tokens;
    int i, len = line.size();
    
    sh->length = len;
    sh->r = 0x00;
    sh->g = 0x00;
    sh->b = 0x00;
    sh->next = NULL;
    
    i = line.find( "//" );
    if( i != string::npos )
    {
        sh->length = i;
        sh->next = new miniAudicle_SyntaxHighlighting;
        sh = sh->next;
        sh->r = 0xff;
        sh->g = 0x00;
        sh->b = 0x00;
        sh->next = NULL;
        sh->length = len - i;
    }
    
    return TRUE;
}

t_CKBOOL miniAudicle::probe()
{
    interfaces.clear();
    
    RtAudio * rta = NULL;
    RtAudioDeviceInfo info;
    
    // allocate RtAudio
    try 
    {
        rta = new RtAudio( );
    }
    catch( RtError & error )
    {
        // problem finding audio devices, most likely
        EM_log( CK_LOG_WARNING, "(RtAudio): %s", error.getMessageString() );
        return FALSE;
    }
    
    // get count    
    int devices = rta->getDeviceCount();
    default_input = devices;
    default_output = devices;
    int first_input = devices, first_output = devices;
    
    // loop
    for( int i = 1; i <= devices; i++ )
    {
        try
        { 
            interfaces.push_back( rta->getDeviceInfo( i ) );
            
/*            if( interfaces[i - 1].inputChannels )
            {
                if( first_input == devices )
                    first_input = i - 1;
                if( default_input == devices && interfaces[i - 1].isDefault )
                    default_input = i - 1;
            }
*/            
            if( interfaces[i - 1].isDefault && interfaces[i - 1].inputChannels 
                && default_input == devices )
                default_input = i - 1;
            
/*            if( interfaces[i - 1].outputChannels )
            {
                if( first_output == devices )
                    first_output = i - 1;
                if( default_output == devices && interfaces[i - 1].isDefault )
                    default_output = i - 1;
            }
*/            
            if( interfaces[i - 1].isDefault && interfaces[i - 1].outputChannels 
                && default_output == devices )
                default_output = i - 1;
        }
        catch( RtError & error )
        {
            EM_log( CK_LOG_WARNING, "(RtAudio): %s", error.getMessageString() );
            break;
        }
    }
    
    if( default_input == devices )
        // no default input found
        default_input = 0;
    
    if( default_output == devices )
        // no default output found
        default_output = 0;
    
    delete rta;
    
    return TRUE;
}

const vector< RtAudioDeviceInfo > & miniAudicle::get_interfaces()
{
    return interfaces;
}

//-----------------------------------------------------------------------------
// name: set_num_inputs()
// desc: set number of virtual machine input channels
//-----------------------------------------------------------------------------
t_CKBOOL miniAudicle::set_num_inputs( t_CKUINT num )
{
    // sanity check
    int max;
    if( interfaces.size() == 0 )
        max = 2;
    else if( vm_options.adc == 0 )
        max = interfaces[default_input].inputChannels;
    else
        max = interfaces[vm_options.adc - 1].inputChannels;
    
    if( num < 0 || num > max )
        vm_options.num_inputs = max;
    else
        vm_options.num_inputs = num;
    
    return TRUE;
}

//-----------------------------------------------------------------------------
// name: get_num_inputs()
// desc: return the number of virtual machine input channels
//-----------------------------------------------------------------------------
t_CKUINT miniAudicle::get_num_inputs()
{
    return vm_options.num_inputs;
}

//-----------------------------------------------------------------------------
// name: set_num_outputs()
// desc: set the number of virtual machine output channels
//-----------------------------------------------------------------------------
t_CKBOOL miniAudicle::set_num_outputs( t_CKUINT num )
{
    // sanity check
    int max;
    if( interfaces.size() == 0 )
        max = 2;
    else if( vm_options.dac == 0 )
        max = interfaces[default_output].outputChannels;
    else
        max = interfaces[vm_options.dac - 1].outputChannels;
    
    if( num < 0 || num > max )
        vm_options.num_outputs = max;
    else
        vm_options.num_outputs = num;

    return TRUE;
}

//-----------------------------------------------------------------------------
// name: get_num_outputs()
// desc: return number of virtual machine output channels
//-----------------------------------------------------------------------------
t_CKUINT miniAudicle::get_num_outputs()
{
    return vm_options.num_outputs;
}

//-----------------------------------------------------------------------------
// name: set_enable_audio()
// desc: specify whether or not to enable audio
//-----------------------------------------------------------------------------
t_CKBOOL miniAudicle::set_enable_audio( t_CKBOOL en )
{
    if( en )
        vm_options.enable_audio = TRUE;
    else
        vm_options.enable_audio = FALSE;
    
    return TRUE;
}

//-----------------------------------------------------------------------------
// name: get_enable_audio()
// desc: determine if audio is enabled
//-----------------------------------------------------------------------------
t_CKBOOL miniAudicle::get_enable_audio()
{
    return vm_options.enable_audio;
}

//-----------------------------------------------------------------------------
// name: set_enable_network_thread()
// desc: specify whether or not to enable the network command thread
//-----------------------------------------------------------------------------
t_CKBOOL miniAudicle::set_enable_network_thread( t_CKBOOL en )
{
    if( en )
        vm_options.enable_network = TRUE;
    else
        vm_options.enable_network = FALSE;
    
    return TRUE;
}

//-----------------------------------------------------------------------------
// name: get_enable_audio()
// desc: determine if the network command thread is enabled
//-----------------------------------------------------------------------------
t_CKBOOL miniAudicle::get_enable_network_thread()
{
    return vm_options.enable_network;
}

t_CKBOOL miniAudicle::set_dac( t_CKUINT dac )
{
    // sanity check
    if( dac < 0 || dac > interfaces.size() )
        vm_options.dac = 0;
    else
        vm_options.dac = dac;
    
    // set parameters to a reasonable value, if necessary
    set_num_outputs( get_num_outputs() );
    
    return TRUE;
}

t_CKUINT miniAudicle::get_dac()
{
    return vm_options.dac;
}

t_CKBOOL miniAudicle::set_adc( t_CKUINT adc )
{
    // sanity check
    if( adc < 0 || adc > interfaces.size() )
        vm_options.adc = 0;
    else
        vm_options.adc = adc;

    // set parameters to a reasonable value, if necessary
    set_num_inputs( get_num_inputs() );
    
    return TRUE;
}

t_CKUINT miniAudicle::get_adc()
{
    return vm_options.adc;
}

t_CKBOOL miniAudicle::set_sample_rate( t_CKUINT srate )
{
    if( interfaces.size() == 0 )
    {
        vm_options.srate = SAMPLING_RATE_DEFAULT;
        return TRUE;
    }
    
    // sanity checks
    // ensure that dac and adc support the given sample rate
    vector< int > & dac_sample_rates = interfaces[( vm_options.dac ? vm_options.dac - 1 : default_output )].sampleRates;
    vector< int >::size_type i, len = dac_sample_rates.size();
    for( i = 0; i < len; i++ )
    {
        if( dac_sample_rates[i] == srate )
            break;
    }
    
    if( i == len )
        // the specified sample rate isnt support by the dac
    {
        vm_options.srate = SAMPLING_RATE_DEFAULT; // hope this one works!
        return TRUE;
    }
    
    vector< int > & adc_sample_rates = interfaces[( vm_options.adc ? vm_options.adc - 1 : default_input )].sampleRates;
    len = adc_sample_rates.size();
    for( i = 0; i < len; i++ )
    {
        if( adc_sample_rates[i] == srate )
            break;
    }
    
    if( i == len )
        // the specified sample rate isnt support by the adc
    {
        vm_options.srate = SAMPLING_RATE_DEFAULT; // hope this one works!
        return TRUE;
    }
    
    vm_options.srate = srate;
    return TRUE;
}

t_CKUINT miniAudicle::get_sample_rate()
{
    return vm_options.srate;
}

t_CKUINT next_power_2( t_CKUINT n )
{
    t_CKUINT nn = n;
    for( ; n &= n-1; nn = n );
    return nn * 2;
}

t_CKBOOL miniAudicle::set_buffer_size( t_CKUINT size )
{
	vm_options.buffer_size = next_power_2( size - 1 );
	return TRUE;
}

t_CKUINT miniAudicle::get_buffer_size()
{
	return vm_options.buffer_size;
}

//-----------------------------------------------------------------------------
// name: get_new_class_names()
// desc: Pull a list of global class names from the type checker, and determine 
//       which ones are new since the last call to this function.  Uses an 
//       existence map to determine the new classes.  Stores the new names in v.
//-----------------------------------------------------------------------------
t_CKBOOL miniAudicle::get_new_class_names( vector< string > & v )
{
    v.clear();
    
    // get global type names from compiler
    vector< Chuck_Type * > types;
    compiler->env->global()->get_types( types );
        
    int i, len = types.size();
    for( i = 0; i < len; i++ )
    {
        if( types[i] )
            if( class_names->insert( pair< string, t_CKINT >( types[i]->name, 1 ) ).second )
                v.push_back( types[i]->name );
    }
    
	return TRUE;
}
