/*----------------------------------------------------------------------------
miniAudicle
GUI to chuck audio programming environment

Copyright (c) 2005 Spencer Salazar.  All rights reserved.
http://chuck.cs.princeton.edu/
http://soundlab.cs.princeton.edu/

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
U.S.A.
-----------------------------------------------------------------------------*/

//-----------------------------------------------------------------------------
// file: mAView.cpp
// desc: ChucK source view
//
// author: Spencer Salazar (ssalazar@cs.princeton.edu)
// date: Summer 2006
//-----------------------------------------------------------------------------

#include "chuck_def.h"

#include "wx/wx.h"
#include "wx/config.h"

#include "mAView.h"
#include "mAParentFrame.h"
#include "mAEvents.h"

#ifdef __WINDOWS_DS__
#include "wxw/icons/add.xpm"
#include "wxw/icons/remove.xpm"
#include "wxw/icons/replace.xpm"
#endif

mAChuckWindow::mAChuckWindow( wxView * v, wxMDIChildFrame * frame, 
							  wxWindowID id, const wxPoint & pos, 
							  const wxSize & size, long style)
	: wxStyledTextCtrl( frame, id, pos, size, style )
{
	view = v;
	open_brackets = close_brackets = 0;
	Connect( wxID_ANY, wxEVT_STC_CHARADDED, 
        (wxObjectEventFunction) (wxEventFunction) wxStaticCastEvent( wxStyledTextEventFunction, ( void (wxEvtHandler::*)( wxStyledTextEvent &) ) &mAChuckWindow::OnCharAdded ) );
}

void mAChuckWindow::OnCharAdded( wxStyledTextEvent & event )
{
	int key = event.GetKey();
	if ( key == _T( '\n' ) )
	{
		int indentation = 0, line = LineFromPosition( GetSelectionStart() );

		if( line > 0)
			indentation = GetLineIndentation( line - 1 );

		while( open_brackets )
		{
			open_brackets--;
			indentation += GetTabWidth();
		}
		
		while( close_brackets )
		{
			close_brackets--;
			indentation -= GetTabWidth();
		}

		if( indentation < 0 )
			indentation = 0;

		SetLineIndentation ( line, indentation );
		GotoPos( GetLineIndentPosition( line ) );
	}

	else if( key == _T( '{' ) )
		open_brackets++;

	else if( key == _T( '}' ) )
	{
		if( open_brackets )
			open_brackets--;
		else
		{
			int indentation = 0, line = LineFromPosition( GetSelectionStart() );
		
			if( GetLineIndentPosition( line ) == GetSelectionStart() - 1 )
				// this is the first non-indent character on this line
			{
				indentation = GetLineIndentation( line ) - GetTabWidth();
				
				if( indentation < 0 )
					indentation = 0;

				SetLineIndentation ( line, indentation );
				GotoPos( GetLineIndentPosition( line ) + 1 );
			}
			else
				close_brackets++;
		}
	}
}


IMPLEMENT_DYNAMIC_CLASS( mAChuckView, wxView )

bool mAChuckView::OnCreate( wxDocument * doc, long WXUNUSED( flags ) )
{
	frame = wxGetApp().CreateChildFrame( doc, this );
	
	wxConfigBase * config = wxConfigBase::Get();

	int width, height;
	frame->GetClientSize( &width, &height );

	window = new mAChuckWindow( this, frame, wxID_ANY, wxPoint(0, 0), 
		wxSize( width, height ), wxSTC_STYLE_LINENUMBER );
	
	int font_size;
#ifdef __LINUX__
	config->Read( _T( "/GUI/Editing/FontSize" ), &font_size, 10 );
#else
	config->Read( _T( "/GUI/Editing/FontSize" ), &font_size, 11 );
#endif
	wxFont font( font_size, wxFONTFAMILY_MODERN, wxNORMAL, wxFONTWEIGHT_NORMAL );
	window->StyleSetFont( wxSTC_STYLE_DEFAULT, font );
	
	window->SetTabWidth( 4 );

    window->SetMarginType( 1, wxSTC_MARGIN_NUMBER );
    window->StyleSetBackground( wxSTC_STYLE_LINENUMBER, wxColour( 212, 208, 200 ) );
    window->StyleSetForeground( wxSTC_STYLE_LINENUMBER, *wxBLACK );
    window->SetMarginWidth( 1, window->TextWidth( wxSTC_STYLE_LINENUMBER, _T( "9999") ) );
	window->SetMarginLeft( 5 );
	// set width to 80 columns
	frame->SetSize( window->TextWidth( wxSTC_STYLE_DEFAULT, _T( "                                                                     ") ),
		window->TextHeight( 1 ) * 32 );
	
	window->StyleClearAll();

	window->SetLexer( wxSTC_LEX_CPP );
	window->StyleSetForeground( wxSTC_C_WORD, *wxBLUE );
	window->SetKeyWords( 0, _T( "int float time dur void same if else while " ) 
							_T( "until for break continue return switch " )
							_T( "class extends public static pure this " )
							_T( "super interface implements protected " )
							_T( "private function fun spork const new now " )
							_T( "true false maybe null NULL me pi samp ms " )
							_T( "second minute hour day week dac adc blackhole " ) );
	window->StyleSetForeground( wxSTC_C_COMMENT, wxColour( 0x60, 0x90, 0x10 ) );
	window->StyleSetForeground( wxSTC_C_COMMENTLINE, wxColour( 0x60, 0x90, 0x10 ) );
	window->StyleSetForeground( wxSTC_C_STRING, wxColour( 0x40, 0x40, 0x40 ) );
	window->StyleSetForeground( wxSTC_C_NUMBER, wxColour( 0xd4, 0x80, 0x10 ) );

	frame->CreateStatusBar();
	frame->SetStatusText( _T( "" ) );

#ifdef __WINDOWS_DS__
	/* In Windows, have a toolbar for each document window. 
	   Linux doesnt support this.  */
	wxToolBar * toolbar = frame->CreateToolBar( wxTB_HORIZONTAL | wxTB_TEXT | wxTB_FLAT );
	wxSize bitmap_size( 32, 32 );
	toolbar->SetToolBitmapSize( bitmap_size );


	wxBitmap add_bitmap( add_xpm );
	toolbar->AddTool( mAID_ADD, _T( "Add Shred" ), add_bitmap );

	toolbar->AddSeparator();

	wxBitmap remove_bitmap( remove_xpm );
	toolbar->AddTool( mAID_REMOVE, _T( "Remove Shred" ), remove_bitmap );

	toolbar->AddSeparator();

	wxBitmap replace_bitmap( replace_xpm );
	toolbar->AddTool( mAID_REPLACE, _T( "Replace Shred" ), replace_bitmap );
	/*
	toolbar->AddSeparator();
	toolbar->AddSeparator();
	toolbar->AddSeparator();
	toolbar->AddSeparator();

	wxBitmap removelast_bitmap( _T( "removelast" ) );
	toolbar->AddTool( mAID_REMOVELAST, _T( "Remove Last" ), 
		removelast_bitmap );

	wxBitmap removeall_bitmap( _T( "removeall" ) );
	toolbar->AddTool( mAID_REMOVEALL, _T( "Remove All" ), 
		removeall_bitmap );
	*/
	toolbar->Realize();
#endif

	docid = ma->allocate_document_id();
	frame->RegisterWithWindowMenu();
	
#ifdef __X__
	// X seems to require a forced resize
	int x, y;
	frame->GetSize( &x, &y );
	frame->SetSize( wxDefaultCoord, wxDefaultCoord, x, y );
#endif
	
	frame->Show( true );
	Activate( true );
	
	return true;
}

void mAChuckView::SetMiniAudicle( miniAudicle * ma )
{
	this->ma = ma;
}

// Handled by wxTextWindow
void mAChuckView::OnDraw( wxDC * WXUNUSED( dc ) )
{
}

void mAChuckView::OnUpdate( wxView * WXUNUSED( sender ), 
						    wxObject * WXUNUSED( hint ) )
{
}

void mAChuckView::OnChangeFilename()
{
	wxView::OnChangeFilename();
	frame->ReregisterWithWindowMenu();
}

bool mAChuckView::OnClose( bool deleteWindow )
{
    ma->free_document_id( docid );
	if( !GetDocument()->Close() )
		return false;
	
	Activate( false );

	frame->DeregisterWithWindowMenu();
	
	if( deleteWindow )
	{
		delete frame;
	}

	return true;
}

void mAChuckView::Add()
{
	std::string code = std::string( window->GetText().ToAscii() ),
		name = std::string( frame->GetTitle().ToAscii() ),
		result;
    t_CKUINT shred_id;

	ma->run_code( code, name, docid, shred_id, result );
	
	wxString status( result.c_str(), wxConvUTF8 );
	
	frame->SetStatusText( status );
}


void mAChuckView::Remove()
{
	std::string result;
    t_CKUINT shred_id;

	ma->remove_code( docid, shred_id, result );
	
	wxString status( result.c_str(), wxConvUTF8 );
	
	frame->SetStatusText( status );
}

void mAChuckView::Replace()
{
	std::string code = std::string( window->GetText().ToAscii() ),
        name = std::string( frame->GetTitle().ToAscii() ),
        result;
    t_CKUINT shred_id;

	ma->replace_code( code, name, docid, shred_id, result );

	wxString status( result.c_str(), wxConvUTF8 );
	
	frame->SetStatusText( status );
}


mADocMDIChildFrame::mADocMDIChildFrame( wxDocument * doc, wxView * view, 
	wxMDIParentFrame * parent, wxWindowID id, const wxString & title, 
	const wxPoint & pos, const wxSize & size, long style, const wxString & name )
	: wxDocMDIChildFrame( doc, view, parent, id, title, pos, size, style, name )
{
	m_menuid = -1;
}

void mADocMDIChildFrame::RegisterWithWindowMenu()
{
	if( m_menuid == -1 )
		m_menuid = wxGetApp().AddWindowMenuItem( this );
}

void mADocMDIChildFrame::DeregisterWithWindowMenu()
{
	if( m_menuid != -1 )
		wxGetApp().RemoveWindowMenuItem( m_menuid );
	m_menuid = -1;
}

void mADocMDIChildFrame::ReregisterWithWindowMenu()
{
	if( m_menuid == -1 )
		RegisterWithWindowMenu();
	wxGetApp().ChangeWindowMenuItem( m_menuid );
}



